#!/usr/bin/env python
'''LocatableFeatures can be added to the sequences. They can be located in
a specific region of the sequence and they can be of a sequence type or
non-sequence kind.
They can be complemented, reversed and sliced.
'''

def _is_sequence(item):
    '''It returns True if the given item is a sequence type.

    A sequence type might be a list, a Seq or a RichSeq or any instance
    with a __len__ method.
    A non-sequence type might be an str or an unicode. These are not
    considered sequences, because they are meant to describe no to use their
    letters as a sequence.
    '''
    if item is None:
        return False
    if isinstance(item, str) or isinstance(item, unicode):
        return False
    try:
        len(item)
        return True
    except TypeError:
        return False

def _check_location(feature, location):
    '''It checks that the feature fits into the space provided by the
    location.
        
    If it not fits it will raise an error.
    '''
    #if the feature is an str we don't mind the legnth, because the
    #str is a label for all the range and it shouldn't be mapped to
    #every residue in the range.
    if not _is_sequence(feature):
        return
    if location is None:
        return
    elif  len(feature) != len(location):
        msg = 'Sequence and Location lengths do not match.'
        raise ValueError(msg)

class LocatableFeature(object):
    '''It represents a Feature that can be added to a sequence.

    Several kind of objects can be a feature: str, RichSeq, Seq, list.
    A RichFeature can be associated with a Location in the sequence.
    If it is associated with a Location and an object with a __len__ method 
    (other than a str) is used, RichFeature will check that the object
    fits into the Location.
    This object is inmutable.
    '''
    def __init__(self, feature, name=None, location=None, parent=None):
        '''The RichFeature constructor.

        Keyword arguments:
        feature -- The actual feature or property stored inside the feature
        name -- The id of the feature (default None)
        location -- The Location associated with the feature (default None)
        parent -- The sequence in which this feature is located (default None)
        '''
        _check_location(feature, location)
        self.__feature = feature
        self.name = name
        self.__location = None
        self.__set_location(location)
        self.__parent = None    #the parent sequence of this feature
        if parent is not None:
            self.parent = parent

    def __set_location(self, location):
        '''It sets the location.'''
        #if a feature is of a sequence type and no location is provided 
        #a default one (starting at 0) will be added.
        if location is None and _is_sequence(self.feature):
            from location import Location
            location = Location(start=0, end=len(self.feature) - 1)
        #we have to inform to the location whose its parent now
        if location is not None:
            location.parent = self
        self.__location = location

    def __get_location(self):
        '''It returns the location'''
        return self.__location

    location = property(__get_location)

    def __get_feature(self):
        '''It returns the feature'''
        return self.__feature

    feature = property(__get_feature)

    def __len__(self):
        '''If there is a Location associated it returns the length, otherwise
        it returns 0.'''
        if self.location is None:
            return 0
        else:
            return len(self.location)

    def __str__(self):
        '''It returns a string represeantation of the feature'''
        toprint = 'name: ' + str(self.name) + '\n'
        toprint += 'location: ' + str(self.location) + '\n'
        toprint += 'feature: ' + str(self.feature) + '\n'
        return toprint

    def complement(self):
        '''It returns a complemented feature.'''
        new_feat = self.feature
        #we need the complemented feature, for that we rely on
        #the capability of the instance that is in the feature slot
        try:
            comp_result = new_feat.complement()
            #if the complement method returns a complemented feature
            #the feature was inmutable
            if type(comp_result) == type(new_feat):
                new_feat = comp_result
        except AttributeError:
            #the feature could not be complemented
            #we return the same feature, without complement it
            pylint = True           #this is a stupid hack just
            pylint = not pylint     #to get pylint happy
        #we also have to complemente the location
        new_loc = self.location.complement()
        return LocatableFeature(new_feat, name=self.name, location=new_loc)

    def __parent_end(self):
        '''If the locatable feature has a parent it returns its end coord,
        otherwise it returns self.location.end, or len(self)..'''
        length = None
        #we get the sequence length
        seq = None
        seq = self.parent
        if seq is not None:
            try:
                length = len(seq)
            except AttributeError:
                forpylint = True     #just to keep pylint happy
                forpylint = not forpylint
        loc = self.location
        if length is None and loc is not None:
            length = loc.end + 1
        if length is None:
            #we asume that the length is the one from this locatablefeature
            length = len(self)
        return length

    def __calculate_feature_index(self, index):
        '''Given an index it calculates the index in the feature coord system'''
        loc = self.location
        #we need the index start, stop and step, wether is a slice or an int
        if isinstance(index, int):
            index_start = index
            index_stop = index + 1
            index_step = 1
        else:
            index_start = index.start
            index_stop = index.stop
            index_step = index.step
        #if the indexes are negative we have to calculate the positive ones
        if (index_start is not None and index_start < 0) or \
           (index_stop is not None and index_stop < 0):
            length = self.__parent_end()
            if index_start < 0:
                index_start = length + index_start
            if index_stop < 0:
                index_stop = length + index_stop
            
        if index_start is None:
            feat_start = None
        else:
            feat_start = index_start - loc.start
        if index_stop is None:
            feat_stop = None
        else:
            feat_stop = index_stop - loc.start
        feat_step = index_step
        if isinstance(index, int):
            return feat_start
        else:
            return slice(feat_start, feat_stop, feat_step)

    def __getitem__(self, index):
        '''it implements the slicing.
        
        It always returns a new LocatableFeature.
        '''
        #the location slicing
        loc = self.location
        if loc is None:
            new_loc = None
        else:
            new_loc = loc[index]
        #the feature itself
        feat = self.feature
        if _is_sequence(feat):
            new_index = self.__calculate_feature_index(index)
            new_feat = feat[new_index]
        else:
            new_feat = feat
        name = self.name
        return LocatableFeature(feature=new_feat, location=new_loc, name=name)

    def __set_parent(self, parent):
        '''It sets the RichSeq that holds this LocatableFeature.'''
        from weakref import ref
        self.__parent = ref(parent) #we do not hold the reference directly to
                                    #avoid circular references
    def __get_parent(self):
        '''It returns the RichSeq that holds this LocatableFeature.'''
        if self.__parent is None:
            return None
        else:
            return self.__parent()
    parent = property(__get_parent, __set_parent)
