#!/usr/bin/env python
'''It tests the RichFeature class.'''

import unittest

from biolib.location import Location
from biolib.locatablefeature import LocatableFeature

class Seq(object):
    '''A faked sequence'''
    def __init__(self, length):
        '''It just has a length.'''
        self.length = length
    def __len__(self):
        '''It returns the length'''
        return self.length

class LocatableFeatureTests(unittest.TestCase):
    '''Tests for the RichFeature'''
    def test_feature_property(self):
        '''It checks that we can set a feature and get it.'''
        #a feature with no location
        item = 'hola'
        feat = LocatableFeature(feature=item)
        assert feat.feature is item

        #a non sequence feature with location
        item = 'hola'
        loc = Location(1, 5)
        feat = LocatableFeature(feature=item, location=loc)
        assert feat.feature is item
        assert feat.location is loc

        #a sequence with no location
        #a location is created
        item = [10, 11, 12, 13]
        feat = LocatableFeature(feature=item)
        assert feat.feature is item
        assert feat.location.start == 0
        assert feat.location.end == 3
        
        #a sequence with location
        item = [10, 11, 12, 13]
        loc = Location(10, 13)
        feat = LocatableFeature(feature=item, location=loc)
        assert feat.feature is item
        assert feat.location is loc
        assert feat.location.start == 10
        assert feat.location.end == 13

    def test_error_feature_location_missmatch(self):
        '''When a feature does not fit into the location an error should be
        raised.
        '''
        #in the init
        item = [10, 11, 12, 13]
        loc = Location(10, 14)
        self.assertRaises(ValueError, LocatableFeature, feature=item, 
                                                                   location=loc)

        
    def test_error_mutable(self):
        '''Features are inmutable.'''
        #in the init
        item = [10, 11, 12, 13]
        loc = Location(10, 14)
        feat = LocatableFeature(feature=item)
        try:
            feat.location = loc
            self.fail('location is inmutable AttributeError expected')
        except AttributeError:
            pass
        try:
            feat.feature = 'hola'
            self.fail('feature is inmutable AttributeError expected')
        except AttributeError:
            pass

    def test_len(self):
        '''It tests the length'''
        item = [10, 11, 12, 13]
        feat = LocatableFeature(feature=item)
        assert len(feat) == 4

        loc = Location(10, 14)
        feat = LocatableFeature(feature='hola', location=loc)
        assert len(feat) == 5
        
        #the feature is non-sequence and there is no location
        feat = LocatableFeature(feature='hola')
        assert len(feat) == 0

    def test_complement(self):
        '''The complement method returns a complemented feature'''
        class Seqmut(object):
            '''This mutable seq class has only test porpouses'''
            def __init__(self):
                '''A faked init, we can set nothing'''
                self.sequence = 'ACTG'
            def complement(self):
                '''It's mutable, so we change the sequence'''
                self.sequence = 'CAGT'

        class Seqnonmut(object):
            '''This inmutable seq class has only test porpouses'''
            def __init__(self, comp=False):
                '''A faked init, we can set nothing'''
                if not comp:
                    self.sequence = 'ACTG'
                else:
                    self.sequence = 'CAGT'
            def complement(self):
                '''It's mutable, so we change the sequence'''
                return Seqnonmut(True)

        #with a mutable feature
        feature = Seqmut()
        assert feature.sequence == 'ACTG'
        loc = Location(0, 3, 1)
        feat = LocatableFeature(feature=feature, location=loc, name='hola')
        new_feat = feat.complement()
        assert new_feat.feature.sequence == 'CAGT'
        assert new_feat.location.start == 0
        assert new_feat.location.end == 3
        assert new_feat.location.strand == -1
        assert new_feat.name == 'hola'

        #with a non-mutable feature
        feature = Seqnonmut()
        assert feature.sequence == 'ACTG'
        loc = Location(0, 3, -1)
        feat = LocatableFeature(feature=feature, location=loc, name='hola')
        new_feat = feat.complement()
        assert new_feat.feature.sequence == 'CAGT'
        assert new_feat.location.start == 0
        assert new_feat.location.end == 3
        assert new_feat.location.strand == 1
        assert new_feat.name == 'hola'

        #with a feature non-complementable
        loc = Location(0, 3)
        feat = LocatableFeature(feature='hola', location=loc, name='hola')
        new_feat = feat.complement()
        assert new_feat.feature == 'hola'
        assert new_feat.location.start == 0
        assert new_feat.location.end == 3
        assert new_feat.location.strand == 0
        assert new_feat.name == 'hola'

    def test_reverse(self):
        '''The reverse method returns a reversed feature'''
                
        #reverse with an inmutable object and no reverse method (e.g. a list)
        loc = Location(0, 3)
        feat = LocatableFeature(feature=[0,1,2,3], location=loc, name='hola')
        new_feat = feat[::-1]
        assert new_feat.feature[0] == 3
        assert new_feat.feature[3] == 0
        assert new_feat.location.start == 0
        assert new_feat.location.end == 3
        assert new_feat.name == 'hola'

        #with a non-reversable object
        loc = Location(0, 3)
        feat = LocatableFeature(feature='hola', location=loc, name='hola')
        new_feat = feat[::-1]
        assert new_feat.feature == 'hola'
        assert new_feat.location.start == 0
        assert new_feat.location.end == 3
        assert new_feat.name == 'hola'

        #reverse inside a larger sequence
        seq = Seq(7)
        loc = Location(1, 4, 1)
        feat = LocatableFeature(feature='hola', location=loc, name='hola',
                                                                    parent=seq)
        new_feat = feat[::-1]
        assert new_feat.feature == 'hola'
        assert new_feat.location.start == 2
        assert new_feat.location.end == 5
        assert new_feat.location.strand == 1
        assert new_feat.name == 'hola'
        #with a list
        #0123456
        # ----
        # 0123
        #  3210
        loc = Location(1, 4, 1)
        feat = LocatableFeature(feature=[0,1,2,3], location=loc, name='hola',
                                                                    parent=seq)
        new_feat = feat[::-1]
        assert new_feat.feature[0] == 3
        assert new_feat.feature[3] == 0
        assert new_feat.location.start == 2
        assert new_feat.location.end == 5
        assert new_feat.name == 'hola'

    def test_feature_slice(self):
        '''We can slice a locatable feature.'''
        #a non-sequence feature with location
        #54321
        #01234
        #  ---
        #  ++*
        #  01  <- new reference
        loc = Location(2, 4, 1)
        feat = LocatableFeature(feature='hola', location=loc, name='hola')
        new_feat = feat[2:4]
        assert new_feat.feature == 'hola'
        assert new_feat.location.start == 0
        assert new_feat.location.end == 1
        assert new_feat.location.strand == 1
        assert new_feat.name == 'hola'
        #with negative index
        new_feat = feat[-3:-1]
        assert new_feat.feature == 'hola'
        assert new_feat.location.start == 0
        assert new_feat.location.end == 1
        assert new_feat.location.strand == 1
        assert new_feat.name == 'hola'

        #a non-sequence feature with no location
        feat = LocatableFeature(feature='hola', name='hola')
        new_feat = feat[2:4]
        assert new_feat.feature == 'hola'
        assert new_feat.location is None
        assert new_feat.name == 'hola'

        #a sequence feature with location
        #54321 <- negative
        #01234
        #  ---
        #  ++*
        #  01   <- new reference
        loc = Location(2, 4, 1)
        item = [2, 3, 4]
        feat = LocatableFeature(feature=item, location=loc, name='hola')
        new_feat = feat[2:4]
        assert new_feat.feature[0] == 2
        assert new_feat.feature[1] == 3
        assert len(new_feat.feature) == 2
        assert new_feat.location.start == 0
        assert new_feat.location.end == 1
        assert new_feat.location.strand == 1
        assert new_feat.name == 'hola'
        #negative index
        new_feat = feat[-3:-1]
        assert new_feat.feature[0] == 2
        assert new_feat.feature[1] == 3
        assert len(new_feat.feature) == 2
        assert new_feat.location.start == 0
        assert new_feat.location.end == 1
        assert new_feat.location.strand == 1
        assert new_feat.name == 'hola'

        #a sequence feature with no location
        #012
        #---
        # +++*
        # 01
        item = [0, 1, 2]
        feat = LocatableFeature(feature=item, name='hola')
        new_feat = feat[1:4]
        assert new_feat.feature[0] == 1
        assert new_feat.feature[1] == 2
        assert len(new_feat.feature) == 2
        assert new_feat.location.start == 0
        assert new_feat.location.end == 1
        assert new_feat.location.strand == 0
        assert new_feat.name == 'hola'

        #test slice outside the feature
        #0123456789
        #   ---
        item = [3, 4, 5]
        loc = Location(start=3, end=5, strand=1)
        feat = LocatableFeature(feature=item, name='hola', location=loc)
        new_feat = feat[:]
        assert new_feat.feature[0] == 3
        assert new_feat.feature[2] == 5
        assert len(new_feat.feature) == 3
        assert new_feat.location.start == 3
        assert new_feat.location.end == 5
        assert new_feat.location.strand == 1
        assert new_feat.name == 'hola'

        #some index Errors
        #0123456789
        #  ---
        loc = Location(2, 4, 1)
        item = [2, 3, 4]
        feat = LocatableFeature(feature=item, location=loc, name='hola')
        try:
            new_feat = feat[6:7]
            self.fail('IndexError expected')
        except IndexError:
            pass
        try:
            new_feat = feat[::2]
            new_feat2 = new_feat[1:5:2]
            self.fail('IndexError expected')
        except IndexError:
            pass

        #test slice using and int index
        #012345
        #   ---
        #    +
        #    0
        item = [3, 4, 5]
        loc = Location(start=3, end=5, strand=1)
        feat = LocatableFeature(feature=item, name='hola', location=loc)
        new_feat = feat[4]
        assert new_feat.feature == 4
        assert new_feat.location.start == 0
        assert new_feat.location.end == 0
        assert new_feat.location.strand == 1
        assert new_feat.name == 'hola'
        #the same with a negative index
        new_feat = feat[-2]
        assert new_feat.feature == 4
        assert new_feat.location.start == 0
        assert new_feat.location.end == 0
        assert new_feat.location.strand == 1
        assert new_feat.name == 'hola'
        #with a negative index in the limit
        #this has been added because of a corner case bug
        new_feat = feat[-1]
        assert new_feat.feature == 5
        assert new_feat.location.start == 0
        assert new_feat.location.end == 0
        assert new_feat.location.strand == 1
        assert new_feat.name == 'hola'

        #inside a bigger sequence
        #7654321
        #0123456
        # ----
        # 0123
        #  ++*
        #  12
        seq = Seq(7)
        loc = Location(1, 4, 1)
        feat = LocatableFeature(feature=[0,1,2,3], location=loc, name='hola',
                                                                    parent=seq)
        new_feat = feat[2:4]
        assert new_feat.feature[0] == 1
        assert new_feat.feature[1] == 2
        assert new_feat.location.start == 0
        assert new_feat.location.end == 1
        assert new_feat.name == 'hola'
        #the same with negative indexes
        new_feat = feat[-5:-3]
        assert new_feat.feature[0] == 1
        assert new_feat.feature[1] == 2
        assert new_feat.location.start == 0
        assert new_feat.location.end == 1
        assert new_feat.name == 'hola'
        #with one negative int as index
        new_feat = feat[-5]
        assert new_feat.feature == 1
        assert new_feat.location.start == 0
        assert new_feat.location.end == 0
        assert new_feat.name == 'hola'

    def test_parent(self):
        '''We can set a parent LocatableFeature for the Location.'''
        loc = Location(1, 100)
        class AClass(object):
            pass
        the_parent = AClass()
        loc.parent = the_parent
        assert loc.parent is the_parent

def main():
    unittest.main()

if __name__ == '__main__':
    main()
