#!/usr/bin/env python
'''It represents a location range in a sequence.
It just has three properties: start, end and strand.
Its not mutable, but it has a complement and a reverse method that return a
new instance.
It also has method to compare it with other locations, like overlaps, contains
and equals that return booleans.
'''

class Location(object):
    '''It represents a location range in a sequence.
    It just has three properties: start, end and strand.
    Its not mutable, but it has a complement and a reverse method that return a
    new instance.
    It also has method to compare it with other locations, like overlaps,
    contains and equals that return booleans.
    '''
    def __init__(self, start, end, strand=0, parent=None, step=1):
        '''The initilizer. Both begin and end are included in the Location.

        keyword arguments:
        start -- The fisrt residue of the range (an int)
        end -- The last residue of the range (an int)
        strand -- the strand (default 0)
        parent -- the LocatableFeature that holds this location (default None)
        step -- annalogous to the slice step (default 1)
        '''
        self.__start = None
        self.__end = None
        self.__strand = None
        self.__step = step     #annalogous to the slice step
        self.__set_start(start)
        self.__set_end(end)
        self.__set_strand(strand)
        self.__parent = None  #a weakref to the LocatableFeature parent
        if parent is not None:
            self.parent = parent

    def __str__(self):
        toprint = 'start:' + str(self.start) + ' '
        toprint += 'end:' + str(self.end) + ' '
        toprint += 'strand:' + str(self.strand) + ' '
        return toprint

    def __set_start(self, start):
        '''It sets the start position of the range.

        It should be 0 or positive.
        keyword arguments:
        start -- The fisrt residue of the range
        '''
        if start >= 0:
            self.__start = int(start)
        else:
            raise ValueError('The Location start position should be ' + \
                             'an integer equal or greater than 0')

    def __get_start(self):
        '''It returns the start position.'''
        return self.__start

    start = property(__get_start)

    def __set_end(self, end):
        '''It sets the end position of the range.
    
        It should be equal or greater than the start
        keyword arguments:
        end -- The last residue of the range
        '''
        if end < self.__start:
            raise ValueError('The Location end position should be an ' + \
                             'integer equal or greater than the start')
        #the end should match an integer numer of times the step
        end = ((end - self.start) // self.__step) * self.__step + self.start
        self.__end = end

    def __get_end(self):
        '''It returns the end position.'''
        return self.__end

    end = property(__get_end)

    def __set_strand(self, strand):
        '''It sets the strand.

        It can be -1, 0 or +1.
        '''
        if strand == 1 or strand == 0 or strand == -1:
            self.__strand = strand
        else:
            raise ValueError('The Location strand should be -1, 0 or +1')

    def __get_strand(self):
        '''It returns the strand'''
        return self.__strand

    strand = property(__get_strand)

    def __get_step(self):
        '''It returns the step.'''
        return self.__step
    step = property(__get_step)

    def __len__(self):
        '''It returns the length of the range.
        
        It takes into account the step.
        '''
        return (self.__end - self.__start) // self.__step + 1

    def __getitem__(self, index):
        '''It implements the slicing mecanism.'''
        #the local variables used
        #           0   1   2   3   4   5   6   7   8   9   0   1
        #location                   -   -   -   -   -   -
        #                           |                   |
        #                           loc_start           loc_end
        #slice              +   +   +   +   +   *
        #                   |               |   |
        #                   istart       iend   istop
        #new_reference      0   1   2   3   4   5   6    7   8   9
        #offset             -   -
        #new_location               -   -   -
        #                           |       |
        #               new_loc_start       new_loc_end

        loc_start = self.start
        loc_end = self.end

        #first we get the start, end and step requested
        if isinstance(index, int):
            istart = index
            iend = index
            istep = 1
        elif isinstance(index, slice):
            istart = index.start
            if istart is None:
                istart = 0
            istop = index.stop
            if istop is None:
                istop = loc_end + 1
            iend = istop - 1
            istep = index.step
            if istep is None:
                istep = 1
        #if there are negative indexes we have to transform them to positive
        if istart < 0 or iend < 0:
            length = self.__parent_length()
            if istart < 0:
                istart = length + istart
            if iend < 0:
                iend = length + iend
        #offset is needed to calculate relative to the new reference
        offset = istart

        #the reverse support
        if istep == -1 and istart == 0 and iend == loc_end:
            return self.__reverse()
        #the other reverse cases are not supported
        if istep == -1:
            msg = 'step -1 not supported with start and stop, use reverse'
            raise IndexError(msg)
        elif istep < -1:
            msg = 'steps lower than -1 not supported'
            raise IndexError(msg)
        #if the step is not 1 or -1 we have to set the new step
        new_step = abs(istep)
        if istep > 1 and self.__step != 1:
            msg = 'two consecutive slices with non 1 steps not supported'
            raise IndexError(msg)
        
        #indexes out of the location errors
        if iend < loc_start or istart > loc_end:
            raise IndexError('Index not covered by the Location')

        #we need to calculate the end, it will be close to the given end
        #but we have to keep in mind the step
        iend = istart + ((iend - istart) // istep) * istep

        #now we take into account that we can not be outside the location
        if istart < loc_start:
            istart = loc_start
        new_start = istart
        if iend > loc_end:
            iend = loc_end
        new_end = iend

        #we need to transform the new_start and new_end into the new
        #reference
        new_start = new_start - offset
        new_end = new_end - offset

        return Location(start=new_start, end=new_end, strand=self.strand,
                           step=new_step)

    def __parent_length(self):
        '''It returns the length of the parent.
        
        If there is no parent it returns the length of this location.
        '''
        length = None
        #we get the sequence length
        loc_feat = self.parent
        seq = None
        if loc_feat is not None:
            try:
                seq = loc_feat.parent
            except AttributeError:
                forpylint = True     #just to keep pylint happy
                forpylint = not forpylint
        if seq is not None:
            try:
                length = len(seq)
            except AttributeError:
                forpylint = True     #just to keep pylint happy
                forpylint = not forpylint
        if length is None:
            #we asume that the length is the one from this biolocation
            length = self.end + 1   #we count from 0
        return length

    def __reverse(self):
        '''It returns a reversed location.'''
        length = self.__parent_length()
        #external_seq 0123456789
        #range         |---|    
        #offset             |--|
        #new_range        |---| 
        offset = length - self.end - 1
        new_end = length - self.start - 1
        new_start = offset
        return Location(new_start, new_end, self.strand)

    def overlaps(self, location):
        '''It returns True if the locations overlap, False otherwise.
        
        Keyword argument:
        location -- A Location to compare.
        '''
        #self    -----     -----     -----     --------
        #range  ---      ---------    ------    -----
        if self.start >= location.start and self.start <= location.end:
            return True
        if self.end >= location.start and self.end <= location.end:
            return True
        if self.start <= location.start and self.end >= location.end:
            return True
        return False

    def contains(self, location):
        '''It returns True if this location totally contains the given location.
        
        Keyword argument:
        location -- A Location to compare.
        '''
        #self  ------------
        #range   --------- 
        if self.start <= location.start and self.end >= location.end:
            return True
        return False

    def equals(self, location):
        '''It returns True if this location is equal to the given location.
        
        Keyword argument:
        location -- A Location to compare.
        '''
        #self  ---------
        #range --------- 
        if self.start == location.start and self.end == location.end:
            return True
        return False

    def complement(self):
        '''It returns a complemented location.
        
        It just changes the strand.
        '''
        new_strand = self.strand
        if new_strand:  #different than 0
            new_strand = -new_strand
        return Location(self.start, self.end, new_strand)

    def __set_parent(self, parent):
        '''It sets the LocatableFeature that holds this Location.'''
        from weakref import ref
        self.__parent = ref(parent) #we do not hold the reference directly to
                                    #avoid circular references
    def __get_parent(self):
        '''It returns the LocatableFeature that holds this Location.'''
        if self.__parent is None:
            return None
        else:
            return self.__parent()
    parent = property(__get_parent, __set_parent)

