#!/usr/bin/env python

import unittest
from location import Location

#a couple of faked classes used for testing
class Seq(object):
    '''A faked sequence'''
    def __init__(self, length):
        '''It just has a length.'''
        self.length = length
    def __len__(self):
        '''It returns the length'''
        return self.length
class LocFeat(object):
    '''A faked locatableFeature'''
    def __init__(self, parent):
        self.parent = parent

class LocationTests(unittest.TestCase):
    '''It tests all the functionallity offered by the Location class.'''
    def test_basic_functionality(self):
        '''It tests the start and end properties and the length'''
        loc = Location(start=0, end=1)
        assert loc.start == 0
        assert loc.end == 1
        assert loc.strand == 0
        assert len(loc) == 2

        loc = Location(start=10, end=20, strand=-1)
        assert  loc.start == 10
        assert loc.end == 20
        assert loc.strand == -1
        assert len(loc) == 11
        
        #if we try to create a biolocation with a step different than 1
        #and and end that does not fit an exact multiple it will fix the end
        loc = Location(start=1, end=7, step=2)
        assert  loc.start == 1
        assert loc.end == 7
        loc = Location(start=1, end=7, step=4)
        assert  loc.start == 1
        assert loc.end == 5

    def test_expections(self):
        '''It checks that some error are raised'''
        self.failUnlessRaises(ValueError, Location, start=1, end=0)
        self.failUnlessRaises(ValueError, Location, start=-1, end=10)
        self.failUnlessRaises(ValueError, Location, start=10, end=20,
                                                                       strand=2)

    def test_overlaps(self):
        '''Locations overlaps'''
        #overlaps
        #   -----
        # ----
        assert Location(start=2, end=5).overlaps(Location(start=1, end=3))
        # ----
        #-------
        assert Location(start=4, end=5).overlaps(Location(start=1, end=6))
        #-----
        # ------
        assert Location(start=1, end=5).overlaps(Location(start=2, end=6))
        #-------
        # ----
        assert Location(start=1, end=5).overlaps(Location(start=2, end=4))
        #-------
        #       ------
        assert not Location(start=0, end=5).overlaps(Location(6, 10))
        #           -------
        #-------
        assert not Location(start=10, end=55).overlaps(Location(6, 8))

        #contains
        #   -----
        # ----
        assert not Location(start=2, end=5).contains(Location(1, 3))
        # ----
        #-------
        assert not Location(start=4, end=5).contains(Location(1, 6))
        #-----
        # ------
        assert not Location(start=1, end=5).contains(Location(2, 6))
        #-------
        # ----
        assert Location(start=1, end=5).contains(Location(2, 4))
        #-------
        #       ------
        assert not Location(start=0, end=5).contains(Location(6, 10))
        #           -------
        #-------
        assert not Location(start=10, end=55).contains(Location(6, 8))

        #equals
        #   -----
        # ----
        assert not Location(start=2, end=5).equals(Location(1, 3))
        # ----
        #-------
        assert not Location(start=4, end=5).equals(Location(1, 6))
        #-----
        # ------
        assert not Location(start=1, end=5).equals(Location(2, 6))
        #-------
        # ----
        assert not Location(start=1, end=5).equals(Location(2, 4))
        #-------
        #       ------
        assert not Location(start=0, end=5).equals(Location(6, 10))
        #           -------
        #-------
        assert not Location(start=10, end=55).equals(Location(6, 8))
        #  --------
        #  --------
        assert Location(start=1, end=5).equals(Location(1, 5))

    def test_reverse(self):
        '''Test that the Location is not mutable, when reversed it returns a new
        instance.'''
        # 012345
        #  |---|
        # |---|
        loc = Location(start=1, end=5, strand=1)
        loc2 = loc[::-1]
        assert loc2.start == 0
        assert loc2.end == 4
        assert loc2.strand == 1

        #reverse with a parent sequence of 10 residues
        seq = Seq(10)
        loc_feat = LocFeat(seq)
        # 0123456789
        #  |---|
        #     |---|
        loc = Location(start=1, end=5, strand=-1, parent=loc_feat)
        loc2 = loc[::-1]
        assert loc2.start == 4
        assert loc2.end == 8
        assert loc2.strand == -1

        # 01234567890
        #      |---|
        #  |---|
        seq = Seq(11)
        loc_feat = LocFeat(seq)
        loc = Location(start=5, end=9, parent=loc_feat)
        loc2 = loc[::-1]
        assert loc2.start == 1
        assert loc2.end == 5
        assert loc2.strand == 0

        try:
            #the lower than -1 strides are not supported
            loc2 = loc[::-2]
            self.fail('IndexError expected')
        except IndexError:
            pass
        try:
            #two consecutive slices with non-1 steps are not supported
            loc2 = loc[::2]
            loc3 = loc2[::2]
            self.fail('IndexError expected')
        except IndexError:
            pass

    def test_complement(self):
        '''Test that the Location is not mutable, when complemented it returns
        a new instance.'''
        loc = Location(start=1, end=5, strand=1)
        loc2 = loc.complement()
        assert loc2.start == 1
        assert loc2.end == 5
        assert loc2.strand == -1

        loc = Location(start=1, end=5, strand=-1)
        loc2 = loc.complement()
        assert loc2.start == 1
        assert loc2.end == 5
        assert loc2.strand == 1

        loc = Location(start=1, end=5)
        loc2 = loc.complement()
        assert loc2.start == 1
        assert loc2.end == 5
        assert loc2.strand == 0

    def test_slices(self):
        '''We can get a sliced location, in a similar way as if it was a
        list.'''
        #a standard slice inside the location
        #54321 <- negative index
        #01234
        #  ---
        #  +++*
        #  012
        loc = Location(start=2, end=4, strand=1)
        loc2 = loc[2:5]
        assert loc2.start == 0
        assert loc2.end == 2
        assert loc2.strand == 1
        #with negative indexes
        loc = Location(start=2, end=4, strand=1)
        loc2 = loc[-3:]
        assert loc2.start == 0
        assert loc2.end == 2
        assert loc2.strand == 1

        #only partially inside the location
        #54321 <- negative index
        #01234567
        #  ---
        # +++*
        # 012
        loc = Location(start=2, end=4, strand=1)
        loc2 = loc[1:4]
        assert loc2.start == 1
        assert loc2.end == 2
        assert loc2.strand == 1
        #with negative indexes
        loc = Location(start=2, end=4, strand=1)
        loc2 = loc[-4:-1]
        assert loc2.start == 1
        assert loc2.end == 2
        assert loc2.strand == 1

        #01234567
        #  ---
        #   +++*
        #   01
        loc = Location(start=2, end=4, strand=1)
        loc2 = loc[3:6]
        assert loc2.start == 0
        assert loc2.end == 1
        assert loc2.strand == 1

        #01234567
        #  ---
        # +++++*
        # 0123
        loc = Location(start=2, end=4, strand=1)
        loc2 = loc[1:6]
        assert loc2.start == 1
        assert loc2.end == 3
        assert loc2.strand == 1

        #an edge case
        #01234567
        #  ---
        #    +++*
        #    0
        loc = Location(start=2, end=4, strand=1)
        loc2 = loc[4:7]
        assert loc2.start == 0
        assert loc2.end == 0
        assert loc2.strand == 1

        #an edge case
        #54321 <- negative index
        #01234
        #  ---
        #++++*
        #0123
        loc = Location(start=2, end=4, strand=1)
        loc2 = loc[:4]
        assert loc2.start == 2
        assert loc2.end == 3
        assert loc2.strand == 1
        #with negative index
        loc2 = loc[-5:-1]
        assert loc2.start == 2
        assert loc2.end == 3
        assert loc2.strand == 1
        
        #a single int as index
        #54321 <- negative index
        #01234
        #  ---
        #   ^
        #   0
        loc = Location(start=2, end=4, strand=1)
        loc2 = loc[3]
        assert loc2.start == 0
        assert loc2.end == 0
        assert loc2.strand == 1
        #with negative index
        loc2 = loc[-2]
        assert loc2.start == 0
        assert loc2.end == 0
        assert loc2.strand == 1

        #None returns everything
        loc = Location(start=10, end=20, strand=1)
        loc2 = loc[:None]
        assert loc2.start == 10
        assert loc2.end == 20
        assert loc2.strand == 1
        loc = Location(start=10, end=20, strand=1)
        loc2 = loc[:]
        assert loc2.start == 10
        assert loc2.end == 20
        assert loc2.strand == 1

        #a stride different than 1
        #01234567890
        #-----------
        # +   + *
        # 01234
        loc = Location(start=0, end=10, strand=1)
        loc2 = loc[1:7:4]
        assert loc2.start == 0
        assert loc2.end == 4
        assert loc2.strand == 1

        #negative index with a lengthier parent sequence
        #a parent sequence of 10 residues
        seq = Seq(10)
        loc_feat = LocFeat(seq)
        # 0987654321 <- negative index
        # 0123456789
        #  -----
        #    -----*
        #    01234 <- new reference
        loc = Location(start=1, end=5, strand=-1, parent=loc_feat)
        loc2 = loc[3:8]
        assert loc2.start == 0
        assert loc2.end == 2
        assert loc2.strand == -1
        #with the negative index
        loc2 = loc[-7:-2]
        assert loc2.start == 0
        assert loc2.end == 2
        assert loc2.strand == -1

        #index errors
        #outside the location
        loc = Location(start=10, end=20, strand=1)
        try:
            loc2 = loc[5:10]
            self.fail('IndexError expected')
        except IndexError:
            pass
        try:
            loc2 = loc[21:27]
            self.fail('IndexError expected')
        except IndexError:
            pass
        try:
            #a list would return an empty list,
            #but a biolocation is not mutable and it has not sense
            #to have an empty location
            loc2 = loc[5:5]
            self.fail('IndexError expected')
        except IndexError:
            pass
        try:
            #the -1 stride is not supported with start and stop
            loc2 = loc[10:20:-1]
            self.fail('IndexError expected')
        except IndexError:
            pass
        #we cannot slice with a step different from 1 if the loc has step
        #different from 1
        loc = Location(start=10, end=20, step=2)
        try:
            #negative indexes are not supported
            loc2 = loc[10:16:2]
            self.fail('NotImplementedError expected')
        except IndexError:
            pass

    def test_parent(self):
        '''We can set a parent LocatableFeature for the Location.'''
        loc = Location(start=1, end=100)
        class AClass(object):
            pass
        the_parent = AClass()
        loc.parent = the_parent
        assert loc.parent is the_parent

    def test_length(self):
        '''The biolocation is capable of returning its length'''
        loc = Location(start=1, end=5, step=1)
        assert len(loc) == 5

        loc = Location(start=1, end=5, step=2)
        assert len(loc) == 3

def main():
    unittest.main()

if __name__ == '__main__':
    main()
