'''A RichSeq represents a sequence and it's capable of holding some features.

It's a non mutable sequence. It adds features to the simple Seq object from
Biopython.
'''
from Bio.Seq import Seq
from Bio.Alphabet import generic_alphabet

class RichSeq(Seq):
    '''RichSeq represents a sequence and its features.

    A sequence can have, besides a sequence, a name and a list of features.
    RichSeq is not mutable.
    '''
    def __init__(self, seq, alphabet=generic_alphabet, name=None,
                 features = None):
        '''The RichSeq init.

        keyword arguments
            seq -- the sequence (Seq or str)
            alphabet -- the sequence alphabet (default generic_alphabet)
            feature -- a list of LocatableFeatures
        '''
        Seq.__init__(self, seq, alphabet=alphabet)
        self.name = name
        #A public list of features is a nice thing, but it has some drawbacks
        #it would be impossible to control which features are allowed to be
        #added, it will also limit the future modification of the implementation
        #for instance, it would be impossible to retrieve the feature by name.
        #Basically it would break the encapsulation.
        self.__features = []     #the features list
        self.__features_index = {}  #the features indexed by name
        self.add_features(features)

    def add_features(self, features):
        '''It adds a list of LocatableFeatures to the RichSeq.'''
        if features is None:
            return
        for feature in features:
            self.add_feature(feature)

    def add_feature(self, feature):
        '''It adds a LocatableFeature to the RichSeq.'''
        seq_len = len(self)
        #we check that the feature is within the sequence range
        loc = feature.location
        if loc is not None:
            #we can't use for this loc.end > seq_len because of the steps
            #different that 1
            if (loc.start + len(loc)) > seq_len:
                msg = 'The feature is outside the range covered by the seq.'
                raise ValueError(msg)
        #is already a feature with that name ?
        name = feature.name
        if name and self.__features_index.has_key(name):
            msg = 'Another feature with the same name was already added.'
            raise ValueError(msg)
        #we have to inform the feature about its parent
        feature.parent = self
        #we add the feature to the list
        self.__features.append(feature)
        #we add the feature name to the index
        if name:
            self.__features_index[name] = len(self.__features) - 1

    def get_feature(self, name):
        '''It returns a feature by its name.'''
        return self.__features[self.__features_index[name]]

    def get_features(self):
        '''A generator function that returns all the features.'''
        for feature in self.__features:
            yield feature

    def __getitem__(self, index):
        '''It returns a new sliced RichSeq.'''
        alphabet = self.alphabet
        name = self.name
        features = []
        for feat in self.get_features():
            features.append(feat[index])
        seq = str(self)[index]
        return RichSeq(seq=seq, alphabet=alphabet, name=name, features=features)

    def __add__(self, other):
        '''It adds another sequence to this sequence.'''
        from location import Location
        from locatablefeature import LocatableFeature
        #the sequence
        new_seq = Seq.__add__(self, other)
        #the features from self
        feats = self.__features[:]
            
        #we add the features from the other sequence
        #we might deal with objects with no features
        #where does the other sequence start, we have to add that to its 
        #features start and stop
        offset = len(self)
        for feat in other.get_features():
            new_feature = feat.feature
            new_name = feat.name
            loc = feat.location
            if loc is not None:
                new_start = loc.start + offset
                new_end = loc.end + offset
                new_strand = loc.strand
                new_step = loc.step
                new_loc = Location(start=new_start, end=new_end,
                            strand=new_strand, step=new_step)
            else:
                new_loc = None
            new_feat = LocatableFeature(feature=new_feature, name=new_name,
                                                            location=new_loc)
            feats.append(new_feat)
        #the new seq name
        if self.name or other.name:
            new_name = str(self.name) + str(other.name)
        else:
            new_name = None
        return RichSeq(seq=str(new_seq), alphabet=new_seq.alphabet,
                name=new_name, features=feats)

