#!/usr/bin/python

import unittest

from biolib.richseq import RichSeq
from biolib.locatablefeature import LocatableFeature
from biolib.location import Location
from Bio.Alphabet import IUPAC

class SequenceTests(unittest.TestCase):
    '''Test for the RichSeq class'''
    def test_seq_creation(self):
        '''We can create a RichSeq with any valid Bio.Seq init.'''
        seq = RichSeq("TCAAAAGGATGCATCATG", IUPAC.unambiguous_dna)
        seq = RichSeq("TCAAAAGGATGCATCATG")

    def test_add_features(self):
        '''We can add LocatableFeatures to the sequence.'''
        #a single locatable feature
        seq = RichSeq("TCAAAAGGATGCATCATG")
        loc = Location(start=2, end=7)
        feat = LocatableFeature(feature='an anotation', location=loc,
                                                                name='annot1')
        seq.add_feature(feat)
        feat3 = LocatableFeature(feature='an anotation', location=loc,
                                                                name='annot2')
        seq.add_feature(feat3)
        feat2 = seq.get_feature('annot1')
        assert feat is feat2
        feat4 = seq.get_feature('annot2')
        assert feat3 is feat4

        #we can add a feature without name
        loc = Location(start=2, end=7)
        feat = LocatableFeature(feature='another anotation', location=loc)
        seq.add_feature(feat)
        assert len(list(seq.get_features())) == 3

        #we can't add a feature outside the range covered by the sequence
        seq = RichSeq("TCAAAA")
        loc = Location(start=2, end=6)
        feat = LocatableFeature(feature='an anotation', location=loc)
        try:
            seq.add_feature(feat)
            self.fail('ValueError expected')
        except ValueError:
            pass

        #we can't add two features with the same name
        seq = RichSeq("TCAAAAGGATGCATCATG")
        feat = LocatableFeature(feature='error', location=loc, name='annot1')
        try:
            seq.add_feature(feat)
            seq.add_feature(feat)
            self.fail('ValueError expected')
        except ValueError:
            pass

        #if we add a sequence feature without location, a default one will be
        #created
        feat = LocatableFeature(feature=[0, 1, 2, 3])
        seq = RichSeq('ATCTGC', features=[feat])
        feat2 = list(seq.get_features())[0]
        assert feat2.location.start == 0

    def test_len(self):
        '''The sequence has a length.'''
        seq = RichSeq('ATCT')
        assert len(seq) == 4

    def test_slice(self):
        '''The slicing of the sequence.'''
        #simple slices
        feat = LocatableFeature([0,1,2,3])
        seq = RichSeq(seq='ATCG', features=[feat])
        new_seq = seq[1]
        assert str(new_seq) == 'T'
        assert list(new_seq.get_features())[0].feature == 1
        new_seq = seq[-1]
        assert str(new_seq[-1]) == 'G'
        assert list(new_seq.get_features())[0].feature == 3
        new_seq = seq[0:2]
        assert str(seq[0:2]) == 'AT'
        feature = list(new_seq.get_features())[0].feature
        assert feature[0] == 0
        assert feature[1] == 1

        #reverse
        feat = LocatableFeature([0,1,2,3])
        seq = RichSeq(seq='ATCG', features=[feat])
        new_seq = seq[::-1]
        assert str(new_seq) == 'GCTA'
        feature = list(new_seq.get_features())[0].feature
        assert feature[0] == 3
        assert feature[3] == 0

        #step 3 for the thrid nucleotide
        feat = LocatableFeature([0,1,2,3,4,5,6,7])
        seq = RichSeq(seq='ATCGATCG', features=[feat])
        new_seq = seq[::3]
        assert str(new_seq) == 'AGC'
        feature = list(new_seq.get_features())[0].feature
        assert feature[0] == 0
        assert feature[1] == 3
        assert feature[2] == 6
        new_seq = seq[1::3]
        assert str(new_seq) == 'TAG'
        feature = list(new_seq.get_features())[0].feature
        assert feature[0] == 1
        assert feature[1] == 4
        assert feature[2] == 7
        new_seq = seq[2::3]
        assert str(new_seq) == 'CT'
        feature = list(new_seq.get_features())[0].feature
        assert feature[0] == 2
        assert feature[1] == 5

    def test_add(self):
        '''Adding to sequences.'''
        feat1 = LocatableFeature([0,1,2,3], name='feat1')
        seq1 = RichSeq(seq='ATCG', features=[feat1])
        feat2 = LocatableFeature([4,5,6,7], name='feat2')
        seq2 = RichSeq(seq='GCTA', features=[feat2])
        
        seq = seq1 + seq2
        assert str(seq) == 'ATCGGCTA'
        feat1 = seq.get_feature('feat1')
        assert len(feat1.feature) == 4
        assert feat1.feature[0] == 0
        assert feat1.location.start == 0
        feat2 = seq.get_feature('feat2')
        assert len(feat2.feature) == 4
        assert feat2.feature[0] == 4
        assert feat2.location.start == 4

        #TODO add a RichSeq + a Seq

        #we can't add sequences with different alphabets
        seq1 = RichSeq(seq='T', alphabet=IUPAC.ambiguous_dna)
        seq2 = RichSeq("T", IUPAC.protein)
        try:
            seq = seq1 + seq2
            self.faild('TypeError expected.')
        except TypeError:
            pass

def main():
    unittest.main()

if __name__ == '__main__':
    main()
